# Copyright 2015 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

NULL_ID = 0x00
HEAP_CNT_ID = 0x01
HEAP_LENGTH_ID = 0x02
PAYLOAD_OFFSET_ID = 0x03
PAYLOAD_LENGTH_ID = 0x04
DESCRIPTOR_ID = 0x05
STREAM_CTRL_ID = 0x06

DESCRIPTOR_NAME_ID = 0x10
DESCRIPTOR_DESCRIPTION_ID = 0x11
DESCRIPTOR_SHAPE_ID = 0x12
DESCRIPTOR_FORMAT_ID = 0x13
DESCRIPTOR_ID_ID = 0x14
DESCRIPTOR_DTYPE_ID = 0x15
